#!/usr/local/bin/perl
#================================================================
#  wiki/index.pl
#  Author: Jerry A. Pipes
#
#  Yet Another Wiki
#  This script allows users to create and edit simple wiki pages 
#  and publish the changes to an RSS feed.
#================================================================
use FindBin;
use CGI;
CGI::ReadParse(*input);
use XML::RSS;
require "yaw.pl";
print "Content-type: text/html\n\n";

#================================================================
#  VARIABLES
#================================================================

# edit wiki.pl for global variables

# current working directory
my $cwd = $FindBin::Bin;
# global page parameters that are inserted into template.html
my %par;

my $mark = "\263";

#================================================================
#  SUBROUTINES
#================================================================

#----------------------------------------------------------------
# emit_code
#----------------------------------------------------------------
sub emit_code {
   my $depth;
   ($code, $depth) = @_;
   my $tags = '';
   while (@code > $depth) {
      $tags .= "</" . (pop @code) . ">\n<p>";
   }
   while (@code < $depth) {
      push @code, $code;
      $tags .= "<$code>";
   }
   if ($code[$#code] ne $code) {
      $tags .= "</$code[$#code]>\n<$code>";  # split with \n
      $code[$#code] = $code;
   }
   return $tags;
}


#----------------------------------------------------------------
# as_anchor
#----------------------------------------------------------------
sub as_anchor {
   -e "$cwd/pages/$_[0]" ? "<a href=\"?$_[0]\">$_[0]<\/a>" : "<a href=\"?$_[0]\">?<\/a>$_[0]";
}


#----------------------------------------------------------------
# in_place_URL
#----------------------------------------------------------------
sub in_place_URL {
   my $ref = $InPlaceUrl[$_[0]];
   $ref =~ s/^(javascript.{30}).*/$1 .../;
   $ref =~ /\.(gif|jpeg|jpg|png)$/i ? "<img src=\"$ref\">" : "<a href=\"$InPlaceUrl[$num]\">$ref<\/a>";
}


#----------------------------------------------------------------
# format_date
#----------------------------------------------------------------
sub format_date {
   my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime($^T - ($_[0] * 24 * 60 * 60));
   my $month = ('January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December')[$mon];
   $year += 1900 if $year < 1900;
   return "$month $mday, $year";
}


#----------------------------------------------------------------
# format_body
#----------------------------------------------------------------
sub format_body {
   local $_ = shift @_;
   @code= ();
   my $body = '';
   s/&/&amp;/g;

   my $hnum = 0;  # initialized for AsHiddenLink
   my $protocol = "(?:http|https|ftp|mailto|file|gopher|telnet|news)";
   foreach (split(/\n/, $_)) {  # -v- do each explicit source row

   $InPlaceUrl=0;
   while (s/\b(javascript):\S.*/$mark$InPlaceUrl$mark/) {
      $InPlaceUrl[$InPlaceUrl++] = $&
   }
   while (s/\b$protocol:[^\s\<\>\[\]\"'\(\)]*[^\s\<\>\[\]"'\(\)\,\.\?]/$mark$InPlaceUrl$mark/) {
      $InPlaceUrl[$InPlaceUrl++] = $&
   }

   # -v- emit_code block-tag section
   $code = '';
   s/^\t+//;  # legacy pages with tab indented lists
   s/^\s*$/<br>/   && ($code = "...");
   /^ /           && ($body .= &emit_code('pre', 1));

   # - new list versions take left-edge multiple * or #
   s/^(\*+)/<li>/  && ($body .= &emit_code('ul', length $1));
   s/^(\#+)/<li>/  && ($body .= &emit_code('ol', length $1));
   # - allow left-edge : version of definition list, non-tabbed, non-greedy
   s/^(:+)(.+?):( +)/<dt>$2<dd>/  && ($body .= &emit_code(DL, length $1));
   # - allow and transform literal "n." ordered lists, left-edge, one level
   s/^(\d+)\./<li>/  && ($body .= &emit_code(OL, 1));

   s/^!!!!//       && ($body .= &emit_code('h4', 1));
   s/^!!!//        && ($body .= &emit_code('h3', 1));
   s/^!!//         && ($body .= &emit_code('h2', 1));
   s/^\"\"//       && ($body .= &emit_code('blockquote', 1));

   $code  || ($body .= &emit_code('', 0));

   # -v- inline tag section
   s/^-----*/<hr>/;

   s/'{3}(.*?)'{3}/<strong>$1<\/strong>/g;
   s/'{2}(.*?)'{2}/<em>$1<\/em>/g;

   #  s/\b[^][]+\b/&as_anchor($&)/geo;  # $& = entire matched string
   s/\[\[([^][]+)\]\]/&as_anchor($1)/geo;
   s/$mark(\d+)$mark/&in_place_URL($1)/geo;

   $body .= "$_<br>\n";
   }  #-- end foreach

   $body .= &emit_code("", 0);
   return $body;
}


#----------------------------------------------------------------
# search
#----------------------------------------------------------------
sub search {

   # define search pattern
   $pattern = $_[0];
   $pattern =~ s/[+?.*()[\]{}|\\]/\\$&/g;
   $pattern = "\\b\\w*($pattern)\\w*\\b";

   # open data directory
   opendir(DIR, "$cwd/pages");
      while (my $file = readdir(DIR)) {
         next if($file =~ /^\./);
         next if($file =~ /\.bak$/);
         # add file to list
         push(@files, $file);
      }
   closedir(DIR);

   # begin search results
   $body = "<strong>Search results for '$_[0]':</strong><br>\n";

   # step through each file
   foreach $file ( @files ) {
      # temporarily disable line separator
      local $/;
      # open file
      open(SEARCH, "$cwd/pages/$file");
      my $contents = <SEARCH>;
      close(SEARCH);
      # if pattern is found in filename or file contents
      if ($file =~ /$pattern/i || $contents =~ /$pattern/i) {
         # increment hit count
         $hits++;
         # escape spaces in file name
         $filelink = $file;
         $filelink =~ s/ /%20/g;
         # append search result to body
         $body .= "<a href=\"?$filelink\">$file</a><br>\n";
      }
   }

   # assign page parameters
   $par{title} = "Search Results";
   $par{page} = "Search Results";
   $par{body} = $body;
   $par{summary} = ($hits || "No") . " pages found out of $#files pages searched.";

   # display page template
   &display_template;
}


#----------------------------------------------------------------
# lock_owner
#----------------------------------------------------------------
sub lock_owner {
   # open lock file
   open(LOCK, "$cwd/pages/$_[0].lock");
   # read IP address from file
   $owner = <LOCK>;
   # close file
   close(LOCK);
   # if user's IP matches return true
   return $owner;
}


#----------------------------------------------------------------
# unlock_file
#----------------------------------------------------------------
sub unlock_file {

   # if this file is locked by the current user
   if(lock_owner($_[0]) =~ /$_[1]/) {
      # delete lock file
      unlink("$cwd/pages/$_[0].lock");
   }
}


#----------------------------------------------------------------
# edit_page
#----------------------------------------------------------------
sub edit_page {

   # lock_file --------------------------------------------------
   sub lock_file {
      # open lock file
      open(LOCK, ">$cwd/pages/$_[0].lock");
      # write IP address to lock file
      print(LOCK $_[1]);
      # close file
      close(LOCK);
   }

   # is_locked --------------------------------------------------
   sub is_locked {
      # return existence of lock file
      return ( -e "$cwd/pages/$_[0].lock" );
   }

   # text of the page
   my $text = "";
   # lock message
   my $lock_message = "- You have this page locked for editing (click Cancel to unlock) -";
   # file owner
   my $owner = lock_owner($_[0]);
   # if page exists
   if ( -e "$cwd/pages/$_[0]" ) {
      # if file is locked
      if ( is_locked($_[0]) ) {
         # if user is not the current owner
         if($owner ne $ENV{'REMOTE_ADDR'}) {
            # disable editing screen
            $disabled = " disabled";
            # set appropriate lock message
            $lock_message = "- File locked for editing by $owner -";
         }
      }
      else {
         # lock file for editing
         lock_file($_[0], $ENV{'REMOTE_ADDR'});
      }
      # temporarily disable line separator
      local $/;
      # open page file
      open(PAGE, "$cwd/pages/$_[0]");
      # read entire file into a single line
      $_ = <PAGE>;
      # close file
      close(PAGE);
      # escape ampersands
      s/&/&amp;/g;
      # escape less-thans
      s/</&lt;/g;
      # escape greater-thans
      s/>/&gt;/g;
      # store text
      $text = $_;
   }

   # assign page parameters
   $par{title} = "Edit &gt; $_[0]";
   $par{page} = $page;
   $par{body} = "<form name=\"editform\" action=\"./\" method=\"post\"><div class=\"yaw-nav\"><input type=\"submit\" value=\"Save\"$disabled><input type=\"button\" value=\"Cancel\" onclick=\"document.forms.editform.save.name = 'cancel'; document.forms.editform.submit();\"></div><input type=\"hidden\" name=\"save\" value=\"$_[0]\"><label class=\"yaw-lock\">$lock_message<br><textarea name=\"text\" rows=\"16\" wrap=\"virtual\"$disabled>$text</textarea></label></form>";

   # display page template
   &display_template;
}


#----------------------------------------------------------------
# update_feed
#----------------------------------------------------------------
sub update_feed {

   # get current GMT
   $now_seconds = time;
   @aday = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat");
   @amon = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");
   ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($now_seconds);
   $mday = sprintf("%02d", $mday);
   $year += 1900;
   $hour = sprintf("%02d", $hour);
   $min = sprintf("%02d", $min);
   $sec = sprintf("%02d", $sec);
   $now = "$aday[$wday], $mday $amon[$mon] $year $hour:$min:$sec GMT";

   # create RSS 2.0 feed instance
   my $rss = new XML::RSS (version => '2.0');

   # parse body for output
   local $body = &format_body($_[1]);

   # if feed file exists
   if( -e "$cwd/index.xml") {
      # read existing feed
      $rss->parsefile("$cwd/index.xml");
   }

   # update channel info
   $rss->channel(title          => "$feed_title",
                 link           => "$feed_link",
                 description    => "$feed_description",
                 pubDate        => "$now",
                 lastBuildDate  => "$now",
                 generator      => "$feed_link",
                 docs           => "http://blogs.law.harvard.edu/tech/rss"
                 );

   # remove oldest item
   pop(@{$rss->{'items'}}) if (@{$rss->{'items'}} == 10);

   # append newest item
   $rss->add_item(author => "$ENV{'REMOTE_ADDR'}",
                  title => "$feed_title: $_[0]",
                  pubDate => "$now",
                  guid => "$now_seconds",
                  description => "$body",
                  link  => "$feed_link?$_[0]"
                  );

   # save feed
   $rss->save("$cwd/index.xml");


}


#----------------------------------------------------------------
# save_page
#----------------------------------------------------------------
sub save_page {

   # if page data file exists, create backup file
   if (open(CUR, "$cwd/pages/$_[0]")) {
      # read page data
      @hold = <CUR>;
      # close file
      close(CUR);
      # open backup file
      open(BAK, ">$cwd/pages/$_[0].bak");
      # write page data
      print(BAK @hold);
      # close file
      close(BAK);
   }

   if (open(CUR, ">$cwd/pages/$_[0]")) {
      # write text to file
      print(CUR $_[1]);
      # close file
      close(CUR);
   }

   # update RSS feed
   &update_feed($_[0], $_[1]);
}


#----------------------------------------------------------------
# get_page
#----------------------------------------------------------------
sub get_page {

   # get page title (default is 'MainPage')
   my $page = ($_[0]) ? $_[0] : "MainPage";
   # unescape spaces in page title
   $page =~ s/%20/ /g;
   # navigation buttons
   my $nav_buttons = "<div class=\"yaw-nav\"><form class=\"yaw-nav\" action=\"./\" method=\"post\"><input type=\"hidden\" name=\"edit\" value=\"$page\"><input type=\"submit\" value=\"Edit\"></form><input class=\"yaw-nav\" type=\"button\" value=\"Help\" onclick=\"document.location.href='?Help'\"></div>";

   # if page exists
   if ( -e "$cwd/pages/$page" ) {
      # temporarily disable line separator
      local $/;
      # open page file
      open(PAGE, "$cwd/pages/$page");
      # read entire file into a single line
      $_ = <PAGE>;
      # close file
      close(PAGE);
      # format body of page
      $body = &format_body($_);
      # format last modified date
      $date = &format_date (-M "$cwd/pages/$page");
   }
   # if page does not exist
   else {
      # display message in body of page
      $body = "$page does not yet exist.<br>Click the <strong>Edit</strong> button to create it.";
   }

   # assign page parameters
   $par{title} = $page;
   $par{page} = $page;
   $par{action} = $nav_buttons;
   $par{body} = $body;
   $par{summary} = "Page last modified on $date" if $date;

   # display page template
   &display_template;
}


#----------------------------------------------------------------
# display_template
#----------------------------------------------------------------
sub display_template {

   # temporarily disable line separator
   local $/;
   # open template file
   open (TEMPLATE, "$cwd/yaw.html");
   # read entire file into a single line
   $_ = <TEMPLATE>;
   # close file
   close (TEMPLATE);
   # replace '$' tags in template with parameters
   s/\$(\w+)/defined($par{$1}) ? $par{$1} : ''/geo;
   # output template
   print;
}


#================================================================
#  MAIN
#================================================================

# if a search is being performed
if($input{'search'}) {
   &search($input{'search'});
}
# if a page is being edited
elsif($input{'edit'}) {
   &edit_page($input{'edit'});
}
# if an edit has been cancelled
elsif($input{'cancel'}) {
   &unlock_file($input{'cancel'}, $ENV{'REMOTE_ADDR'});
   &get_page($input{'cancel'});
}
# if a page is being saved
elsif($input{'save'}) {
   &save_page($input{'save'}, $input{'text'});
   &unlock_file($input{'save'}, $ENV{'REMOTE_ADDR'});
   &get_page($input{'save'});
}
# otherwise display the requested page
else {
   &get_page($ENV{'QUERY_STRING'});
}


